#!/usr/bin/python3
#
# Copyright (c) Daniel Sheffield 2021
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY

try:
    import psycopg2
    MOCK = False
except:
    from faker import Faker
    Faker.seed(4321)
    fake = Faker()
    MOCK = True
    # first, import a similar Provider or use the default one
    from faker.providers import BaseProvider

    # create new provider class
    class Products(BaseProvider):
        PRODUCTS = {
            'Dark Chocolate': 'Chocolate',
            'Milk Chocolate': 'Chocolate',
            'Cooking Chocolate': 'Chocolate',
            'Cucumber': 'Veggies',
            'Bananas': 'Fruit',
            'Milk': 'Milk and Cream',
            'Cream': 'Milk and Cream',
            'Coffee Beans': 'Coffee',
            'Freerange Eggs': 'Eggs',
        }
        GROUPS = [
            'Fish, Meat, Eggs',
            'Dairy',
            'Beverages',
            'Treats',
            'Produce',
        ]
        CATEGORIES = {
            'Eggs': 'Fish, Meat, Eggs',
            'Milk and Cream': 'Dairy',
            'Chocolate': 'Treats',
            'Veggies': 'Produce',
            'Fruit': 'Produce',
            'Coffee': 'Beverages',
        }
        STORES = [
            'Paknsave',
            'Countdown',
            'New World',
            'Dreamview',
            'TOFS',
        ]
        CATEGORY_UNITS = {
            'Eggs': ('g',),
            'Milk and Cream': ('mL', 'L',),
            'Chocolate': ('g', 'kg',),
            'Veggies': ('g', 'kg', 'Piece',),
            'Fruit': ('g', 'kg', 'Piece', 'Bunch',),
            'Coffee': ('g', 'kg',),
        }
        ORGANIC = [
            True,
            False,
        ]
        def _dict_choice(self, _dict):
            key = fake.random.choice([ i for i in _dict ])
            return key, _dict[key]
        
        def _product_choice(self):
            key = None
            product = []
            for _dict in (
                self.PRODUCTS,
                self.CATEGORIES,
                dict([
                    (k, None) for k in self.GROUPS
                ]),
            ):
                if key is None:
                    k, key = self._dict_choice(_dict)
                    product.append(k)
                    continue
                product.append(key)
                key = _dict[key]
            return product
        
        def product(self):
            return self._product_choice()
        
        def product_unit(self, product):
            return fake.random.choice(self.CATEGORY_UNITS[product[1]])
        
        def description(self, *args, **kwargs):
            return fake.sentence(*args, **kwargs)
            
        def store(self):
            return fake.random.choice(self.STORES)
        
        def unit(self):
            return fake.random.choice(self.UNITS)
        
        def organic(self):
            return fake.random.choice(self.ORGANIC)

    # then add new provider to faker instance
    fake.add_provider(Products)

import urwid
from urwid import numedit
from decimal import Decimal
import time
import datetime
import itertools
from collections import (
    OrderedDict,
)

COPYRIGHT = "Copyright (c) Daniel Sheffield 2021"

class mock_conn(object):
    
    def close(self):
        pass

class mock_cur(object):
    def __init__(self, records):
        self.records = records
    
    def execute(self, *args, **kwargs):
        pass
    
    def fetchall(self):
        yield from self.records
    
    def close(self):
        pass


if MOCK:
    records = [{
        'ts': datetime.datetime(2021, 8, 29, 14, 8, 0, 0),
        'store': 'Countdown',
        'price': 4.30,
        'quantity': 250.0,
        'unit': 'g',
        'organic': False,
        'description': 'Whittakers',
        'product': 'Dark Chocolate',
        'group': 'Treats',
        'category': 'Chocolate',
    }]
    for i in range(0,100):
        product = fake.product()
        quantity = fake.random.random()*500 + 1
        price = fake.random.random()*10
        unit = fake.product_unit(product)
        organic = fake.organic()
        words = []
        words.extend(fake.sentence().split())
        words.append(
            f'{quantity:.2f} {unit} @ {price:.2f}'
        )
        if organic:
            words.append('organic')
        
        records.append({
            'ts': datetime.datetime(2021, 8, 29, 14, 8, 0, 0),
            'store': fake.store(),
            'price': price,
            'quantity': quantity,
            'unit': unit,
            'organic': organic,
            'description': ' '.join([
                product[0],
                *fake.description(ext_word_list=words).split()
            ]),
            'product': product[0],
            'group': product[2],
            'category': product[1],
        })
    col_idx_map = dict([ ( k, idx ) for idx,k in enumerate(sorted(records[0].keys())) ])
    records = [
        [ r[k] for k in sorted(r.keys()) ] for r in records
    ]
    conn = mock_conn()
    cur = mock_cur(records)
else:
    import os
    user = os.getenv('USER')
    conn = psycopg2.connect(f"dbname=das user={user}")
    cur = conn.cursor()

palette = [
    ('banner', 'light gray', 'dark red'),
    ('streak', 'light red', 'dark gray'),
    ('bg', 'light red', 'black'),
]

grid_layout = [
    [ 'ts',          'store',    ],
    [ 'organic',     'product',  ],
    [ 'category',    'group',   ],
]
side_pane = [
    'unit',
    'quantity',
    'price',
]
bottom_pane = [
    'description',
]

cols = [
    c for c in filter(
        lambda x: x is not None,
        itertools.chain(
            *grid_layout, side_pane, bottom_pane
        )
    )
]

#cols.remove(None)

NON_IDENTIFIER_COLUMNS = [
    'ts',
    'store',
    'quantity',
    'unit',
    'price',
    'organic',
]

display_map = {
    'ts': lambda x: f"{time.strftime('%Y-%m-%d %H:%M', (x.year, x.month, x.day, x.hour, x.minute, 0, 0, 0, 0))}",
    'price': lambda x: f"{x:.2f}",
    'quantity': lambda x: f"{x:.2f}",
    'organic': lambda x: "true" if x else "false",
}
display = lambda data, name: display_map[name](data) if name in display_map else data

cur.execute("SELECT * FROM transaction_view;")
if not MOCK:
    col_idx_map = dict([ (d.name, i) for i,d in enumerate(cur.description) if d.name in cols ])

def records(cursor, col_idx_map):
    for row in cursor.fetchall():
        yield dict([
            (name, display(row[i], name)) for name, i in col_idx_map.items()
        ])
    cur.execute("SELECT * FROM transaction_view;")


def record_matches(record, strict=None, **kwargs):
    strict = strict or []
    for k,v in kwargs.items():
        if not v:
            continue
        
        if k in strict and v.lower() != record[k].lower():
            return False
        
        if v.lower() not in record[k].lower():
            return False
        
    return True

def unique_suggestions(name, exclude=NON_IDENTIFIER_COLUMNS, **kwargs):
    exclude = filter(
        lambda x: x != name or name == 'ts',
        exclude,
    )
    [ kwargs.pop(k) for k in exclude if k in kwargs]
    items = suggestions(name, exclude=exclude, **kwargs)
    return sorted(set(map(lambda x: x[name], items)))

def suggestions(name, exclude=NON_IDENTIFIER_COLUMNS, **kwargs):
    exclude = filter(
        lambda x: x != name or name == 'ts',
        exclude,
    )
    [ kwargs.pop(k) for k in exclude if k in kwargs]
    yield from filter(lambda x: record_matches(
        x, strict=[ k for k in kwargs if k != name ], **kwargs
    ), records(cur, col_idx_map))

def show_or_exit(key):
    if isinstance(key, tuple):
        return
    
    if key in ('esc',):
        raise urwid.ExitMainLoop()

def interleave(_list, div):
    for element in _list:
        yield element
        yield div

class AutoCompleteEdit(urwid.Edit):
    def __init__(self, name, *args, apply_change_func=None, **kwargs):
        if isinstance(name, tuple):
            pallete, title = name
            self.name = title
            title = title.title()
            passthrough = (pallete, u'  ' if title.lower() == 'unit' else u'')
        else:
            self.name = name
            title = name.title()
            passthrough = u'  ' if name.lower() == 'unit' else u''
            
        super(AutoCompleteEdit, self).__init__(passthrough, *args, **kwargs)
        self.apply = apply_change_func

    def keypress(self, size, key):
        if key == 'enter':
            self.apply(self.name)
            return
        elif key == 'delete':
            self.set_edit_text('')
            return
        
        return super(AutoCompleteEdit, self).keypress(size, key)

        self.apply(self.name)

class AutoCompleteFloatEdit(numedit.FloatEdit):
    def __init__(self, name, *args, apply_change_func=None, **kwargs):
        self.last_val = None
        self.op = '='
        self.pallete = None
        if isinstance(name, tuple):
            self.pallete, self.name = name
            title = self.name.title()
            passthrough = (self.pallete, f'{self.op} ')
        else:
            self.name = name
            title = name.title()
            passthrough = title
            
        super(AutoCompleteFloatEdit, self).__init__(passthrough, *args, **kwargs)
        self.apply = apply_change_func

    def update_caption(self):
        if self.pallete is not None:
            self.set_caption((self.pallete, f'{self.op} '))
        else:
            self.set_caption(f'{self.op} ')

    def set_op(self, op):
        self.op = op
        self.last_val = self.value()
        self.set_edit_text('')
        self.update_caption()
        
    def calc(self):
        x = self.last_val
        op = self.op
        if op in ('+', '-',):
            y = self.value() or Decimal(0.0)
            if op == '+':
                z = x + y
            else:
                z = x - y
        elif op in ('*', '/'):
            y = self.value() or Decimal(1.0)
            if op == '*':
                z = x * y
            else:
                z = x / y
        else:
            y = self.value() or Decimal(0.0)
            z = y
        
        self.op = '='
        self.update_caption()
        self.set_edit_text(f'{z:.2f}')

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        ops = ('+', '-', '*', '/',)

        if key in ops:
            if self.op in ops:
                self.calc()
            self.set_op(key)
            return
        elif key == 'enter':
            if self.get_edit_text() == '' or self.value() == Decimal(0.0):
                return self.apply(self.name)
            self.calc()
            return
        elif key == '=':
            self.calc()
            return
        elif key == 'delete':
            self.set_edit_text('')
            self.op = '='
            self.update_caption()
            return

        return super(AutoCompleteFloatEdit, self).keypress(size, key)

class NoTabCheckBox(urwid.CheckBox):
    def keypress(self, size, key):
        if not isinstance(key, tuple) and key == 'tab':
            return 
        else:
            return super(NoTabCheckBox, self).keypress(size, key)

def _set_focus_path(container, path):
    try:
        container.set_focus_path(path)
        return
    except IndexError:
        pass
        
    if path[-1] == 0 and len(path) > 1:
        _set_focus_path(container, path[:-1])
        return
    
    raise IndexError

class GroceryTransactionEditor(urwid.WidgetPlaceholder):
    def __init__(self, fields):
        super(GroceryTransactionEditor, self).__init__(urwid.SolidFill(u'/'))
        self.organic_checkbox = NoTabCheckBox(
            u"Organic",
            on_state_change=self.apply_organic_state
        )
        self._init_data(fields)
        self.widgets = dict()
        self.show('transaction')

    def iter_focus_paths(self):
        initial = [2,0,0,0]
        container = self.original_widget.original_widget.original_widget
        _set_focus_path(container, initial)
        while True:
            path = container.get_focus_path()
            yield path
            self.advance_focus()
            path = container.get_focus_path()
            if path == initial:
                self.advance_focus()
                break

    def advance_focus(self, reverse=False):
        container = self.original_widget.original_widget.original_widget
        path = container.get_focus_path()
        if reverse:
            paths = [ i for i in self.iter_focus_paths() ]
            zipped_paths = zip(paths, [
                *paths[1:], paths[0]
            ])
            prev_path = map(lambda x: x[0], filter(
                lambda x: x[1] == path,
                zipped_paths
            ))
            _set_focus_path(container, next(prev_path))
            return
        
        _iter = [ i for i in enumerate(path) ][::-1]

        for idx, part in _iter:
            p = [ i for i in path ]
            if reverse:
                p[idx] -= 1
            else:
                p[idx] += 1
            
            try:
                _set_focus_path(container, p)
                if path == [3]:
                    self.advance_focus(reverse=reverse)
                return
            except IndexError:
                path[idx] = 0
        
        container.set_focus_path([2,0,0,0])

        
    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        
        if key == 'tab':
            self.advance_focus()
        elif key == 'shift tab':
            self.advance_focus(reverse=True)
        else:
            return super(GroceryTransactionEditor, self).keypress(size, key)
        

    def _init_data(self, fields):
        self.data = OrderedDict([
            (k, '') for k in fields
        ])
        self.clear()

    def get_activity(self, name):
        return getattr(self, name)

    def show(self, name, *args, **kwargs):
        if name in self.widgets:
            self.original_widget = self.widgets[name]
            self.update(name)
            return
        
        a = self.get_activity(name)
        widget = a(*args, **kwargs)
        if name != 'suggestions':
            self.widgets[name] = widget
        
        self.original_widget = widget

    def _apply_choice(self, name, value):
        self.data.update({
            name: value,
        })
        for k,v in self.data.items():
            if k == name or v:
                continue
            options = unique_suggestions(k, **self.data)
            if len(options) == 1 and k != 'ts':
                self.data.update({
                    k: list(options)[0],
                })
        self.show('transaction')
    
    def apply_choice(self, name):
        apply = lambda w,x: self._apply_choice(name, x)
        return apply

    def _apply_changes(self, name, value):
        self.data.update({
            name: value,
        })

    def _autocomplete(self, name):
        options = unique_suggestions(name, **self.data)
        if 0 < len(options):
            self.show('suggestions', name, options=options)
    
    def apply_organic_state(self, w, state):
        self.data['organic'] = repr(state).lower()

    def apply_changes(self, name):
        apply = lambda w,x: self._apply_changes(name, x)
        return apply

    def suggestions(self, name, options=None):
        body = [urwid.Text(name.title()), urwid.Divider()]
        for c in options:
            button = urwid.Button(c)
            urwid.connect_signal(button, 'click', self.apply_choice(name), c)
            body.append(urwid.AttrMap(button, None, focus_map='reversed'))
        walker = urwid.SimpleFocusListWalker(body, wrap_around=False)
        listbox = urwid.ListBox(walker)
        pad = urwid.Padding(listbox, left=2, right=2)
        top = urwid.Overlay(pad, self.original_widget,
            align='center', width=('relative', 60),
            valign='middle', height=('relative', 60),
            min_width=20, min_height=9)
        def keypress(size, key, original=top.keypress):
            if key != 'esc':
                return original(size, key)
            self.show('transaction')
        top.keypress = keypress
        return urwid.AttrMap(top, 'banner')
    
    def save(self):
        fmt = '%Y-%m-%dT%H%M'
        ts = self.data['ts']
        ts_log = time.strptime(ts, '%Y-%m-%d %H:%M')
        store = self.data['store']
        description = self.data['description']
        quantity = self.data['quantity']
        unit = self.data['unit']
        price = self.data['price']
        product = self.data['product']
        organic = self.data['organic']
        log = f"{time.strftime(fmt, ts_log)}-{store}.txn"
        with open(log, 'a') as f:
            f.write(
                f"CALL insert_transaction('{ts}', $store${store}$store$, "
                f"$descr${description}$descr$, {quantity}, $unit${unit}$unit$, "
                f"{price}, $produ${product}$produ$, {organic});\n"
            )

    def clear(self):
        for k in self.data:
            if k in ('ts', 'store', 'organic'):
                continue
            self.data[k] = ''
        self.organic_checkbox.set_state(False)
        self.apply_organic_state(self.organic_checkbox,False)

    def save_and_clear(self):
        self.save()
        self.clear()
        self.show('transaction')

    def update(self, name):
        getattr(self, f"update_{name}")()
    
    def update_transaction(self):
        for k in self.edit_fields:
            self.edit_fields[k].set_edit_text(self.data[k])
        self.organic_checkbox.set_state(True if self.data['organic'] == 'true' else False)

    def transaction(self):
        self.edit_fields = OrderedDict()
        for k in self.data:
            if k in side_pane and k != 'unit':
                ef = AutoCompleteFloatEdit(('bg', k), apply_change_func=self._autocomplete)
            elif k != 'organic':
                ef = AutoCompleteEdit(('bg', k), apply_change_func=self._autocomplete)
            else:
                continue
            ef.set_edit_text(self.data[k])
            urwid.connect_signal(ef, 'change', self.apply_changes(k))
            self.edit_fields[k] = ef
    
        header = urwid.Text(u'Fill Transaction', 'center')
        _copyright = urwid.Text(COPYRIGHT, 'center')
        button = urwid.Button(('streak', u'Done'))
        urwid.connect_signal(button, 'click', lambda w: self.save_and_clear())
        banner = urwid.Pile([
            urwid.Padding(header, 'center', width=('relative', 100)),
            urwid.Padding(_copyright, 'center', width=('relative', 100)),
        ])
        banner = urwid.AttrMap(banner, 'banner')
        fields = dict([
            (k, urwid.LineBox(urwid.AttrMap(self.edit_fields[k], 'streak'), title=k.title(), title_align='left')) for k in self.edit_fields
        ])
        side_pane_widget = (12, urwid.Pile([
            fields[r] if r is not None else urwid.Divider() for r in side_pane
        ]))
        main_pane_widgets = []
        for i, r in enumerate(grid_layout):
            widgets = []
            for c in r:
                if c is not None:
                    if c != 'organic':
                        widgets.append(fields[c])
                    else:
                        widgets.append(urwid.LineBox(urwid.AttrMap(self.organic_checkbox, 'bg')))
                else:
                    widgets.append(urwid.Divider())
            main_pane_widgets.append(urwid.Columns(widgets))


        main_pane_widget = urwid.Pile(main_pane_widgets)
        
        widget = urwid.Pile([
            banner,
            urwid.Divider(),
            urwid.Columns((main_pane_widget, side_pane_widget),
                dividechars=2,
            ),
            *[ fields[c] if c is not None else urwid.Divider() for c in bottom_pane ],
            urwid.Divider(),
            button,
        ])
        widget = urwid.Filler(widget, 'top') 
        widget = urwid.AttrMap(widget, 'bg')
        return widget

#screen = urwid.raw_display.Screen()
#screen.set_terminal_properties(colors=256, has_underline=True)
#screen.register_palette(palette)

app = GroceryTransactionEditor(cols)
loop = urwid.MainLoop(app, palette, unhandled_input=show_or_exit)
loop.run()

cur.close()
conn.close()
